using ShellProgressBar;
using System.Text.Json;
using View_by_Distance.Distance.Models.Stateless;
using View_by_Distance.FaceRecognitionDotNet;
using View_by_Distance.Shared.Models;

namespace View_by_Distance.Distance.Models;

public partial class E_Distance
{

    private readonly List<string> _Moved;
    private readonly List<double?> _Debug;
    private readonly List<string> _Renamed;
    private readonly Serilog.ILogger? _Log;
    private readonly int _FaceConfidencePercent;
    private readonly bool _DistanceRenameToMatch;
    private readonly double[] _RangeFaceConfidence;
    private readonly bool _DistanceMoveUnableToMatch;
    private readonly List<string> _AllMappedFaceFiles;
    private readonly double[] _RangeDistanceTolerance;
    private readonly List<string> _AllMappedFaceFileNames;
    private readonly List<string> _DuplicateMappedFaceFiles;

    public E_Distance(bool distanceMoveUnableToMatch, bool distanceRenameToMatch, int faceConfidencePercent, double[] rangeDistanceTolerance, double[] rangeFaceConfidence)
    {
        _Debug = new();
        _Moved = new();
        _Renamed = new();
        _AllMappedFaceFiles = new();
        _AllMappedFaceFileNames = new();
        _DuplicateMappedFaceFiles = new();
        _RangeFaceConfidence = rangeFaceConfidence;
        _Log = Serilog.Log.ForContext<E_Distance>();
        _DistanceRenameToMatch = distanceRenameToMatch;
        _FaceConfidencePercent = faceConfidencePercent;
        _RangeDistanceTolerance = rangeDistanceTolerance;
        _DistanceMoveUnableToMatch = distanceMoveUnableToMatch;
    }

    private void MoveUnableToMatch(string eDistanceContentDirectory, string mappedFaceFile, string mappedFaceFileName)
    {
        bool check;
        string? directoryName = Path.GetDirectoryName(mappedFaceFile);
        if (mappedFaceFileName is null || directoryName is null)
            check = false;
        else
        {
            if (string.IsNullOrEmpty(directoryName) || string.IsNullOrEmpty(directoryName) || !directoryName.Contains(eDistanceContentDirectory))
                check = false;
            else
            {
                List<string> directoryNames = new();
                string? checkDirectoryName = directoryName;
                for (int i = 0; i < int.MaxValue; i++)
                {
                    if (string.IsNullOrEmpty(checkDirectoryName))
                        continue;
                    directoryNames.Add(Path.GetFileName(checkDirectoryName));
                    checkDirectoryName = Path.GetDirectoryName(checkDirectoryName);
                    if (string.IsNullOrEmpty(checkDirectoryName))
                        continue;
                    if (checkDirectoryName == eDistanceContentDirectory)
                        break;
                }
                if (string.IsNullOrEmpty(checkDirectoryName) || !directoryNames.Any() || !long.TryParse(directoryNames[^1][1..^1], out long directoryTicks))
                {
                    check = false;
                    File.Delete(mappedFaceFile);
                }
                else
                {
                    checkDirectoryName = Path.Combine(checkDirectoryName, $"({directoryTicks}_{string.Join('-', _RangeDistanceTolerance)})");
                    for (int i = directoryNames.Count - 1 - 1; i > -1; i--)
                        checkDirectoryName = Path.Combine(checkDirectoryName, directoryNames[i]);
                    if (!Directory.Exists(checkDirectoryName))
                        _ = Directory.CreateDirectory(checkDirectoryName);
                    File.Move(mappedFaceFile, Path.Combine(checkDirectoryName, mappedFaceFileName));
                    check = true;
                }
            }
        }
        if (check)
            _Moved.Add(mappedFaceFile);
    }

    private FaceDistanceContainer[] GetFaceDistanceContainers(MappingFromItem mappingFromItem, List<Face> filteredFaces)
    {
        FaceDistanceContainer[] results;
        int confidencePercent;
        int normalizedRectangle;
        FaceDistance faceDistance;
        FaceDistanceContainer faceDistanceContainer;
        List<FaceDistanceContainer> collection = new();
        foreach (Face face in filteredFaces)
        {
            if (face.FaceEncoding is null || face.Location is null || face.OutputResolution is null)
                throw new NotSupportedException();
            confidencePercent = Shared.Models.Stateless.Methods.ILocation.GetConfidencePercent(_FaceConfidencePercent, _RangeFaceConfidence, face.Location.Confidence);
            normalizedRectangle = Shared.Models.Stateless.Methods.ILocation.GetNormalizedRectangle(face.Location, Shared.Models.Stateless.ILocation.Digits, face.OutputResolution);
            if (face.FaceDistance?.Encoding is not null && face.FaceDistance.Encoding is FaceRecognitionDotNet.FaceEncoding faceEncoding)
                faceDistance = new(confidencePercent, faceEncoding, mappingFromItem.Id, mappingFromItem.IsWrongYear, mappingFromItem.MinimumDateTime, normalizedRectangle);
            else
            {
                faceEncoding = FaceRecognition.LoadFaceEncoding(face.FaceEncoding.RawEncoding);
                faceDistance = new(confidencePercent, faceEncoding, mappingFromItem.Id, mappingFromItem.IsWrongYear, mappingFromItem.MinimumDateTime, normalizedRectangle);
                lock (filteredFaces)
                    face.SetFaceDistance(faceDistance);
            }
            faceDistanceContainer = new(face, faceDistance);
            collection.Add(faceDistanceContainer);
        }
        results = collection.ToArray();
        return results;
    }

    private static List<FaceDistance> GetFaceDistanceEncodings(FaceDistanceContainer[] faceDistanceContainers)
    {
        List<FaceDistance> faceDistanceEncodings = new();
        foreach (FaceDistanceContainer faceDistanceContainer in faceDistanceContainers)
        {
            if (faceDistanceContainer.FaceDistance.Encoding is null)
                continue;
            faceDistanceEncodings.Add(faceDistanceContainer.FaceDistance);
        }
        return faceDistanceEncodings;
    }

    private List<(Face Face, double? Length)> GetValues(MappingFromItem mappingFromItem, Face[] faces, Shared.Models.FaceEncoding modelsFaceEncoding, int normalizedRectangle)
    {
        List<(Face Face, double? Length)> results = new();
        Face face;
        FaceDistance faceDistanceLength;
        List<Face> filteredFaces = FilterByIntersect(faces, normalizedRectangle);
        if (filteredFaces.Any())
        {
            FaceRecognitionDotNet.FaceEncoding faceRecognitionDotNetFaceEncoding = FaceRecognition.LoadFaceEncoding(modelsFaceEncoding.RawEncoding);
            FaceDistance faceDistanceEncoding = new(faceRecognitionDotNetFaceEncoding);
            FaceDistanceContainer[] faceDistanceContainers = GetFaceDistanceContainers(mappingFromItem, filteredFaces);
            int faceDistanceContainersLength = faceDistanceContainers.Length;
            if (faceDistanceContainersLength != filteredFaces.Count)
                throw new NotSupportedException();
            List<FaceDistance> faceDistanceEncodings = GetFaceDistanceEncodings(faceDistanceContainers);
            if (faceDistanceEncodings.Count != filteredFaces.Count)
                throw new NotSupportedException();
            List<FaceDistance> faceDistanceLengths = FaceRecognition.FaceDistances(faceDistanceEncodings, faceDistanceEncoding);
            if (faceDistanceLengths.Count != faceDistanceContainersLength)
                throw new NotSupportedException();
            for (int i = 0; i < filteredFaces.Count; i++)
            {
                face = filteredFaces[i];
                faceDistanceLength = faceDistanceLengths[i];
                if (faceDistanceLength.Length is null)
                    throw new NotSupportedException();
                results.Add(new(face, faceDistanceLength.Length.Value));
            }
        }
        return results;
    }

    private (Face, double?)[] GetClosestFaceByDistanceIgnoringTolerance(MappingFromItem mappingFromItem, int normalizedRectangle, Face[] filteredFaces, Shared.Models.FaceEncoding modelsFaceEncoding)
    {
        (Face, double?)[] results;
        List<(Face Face, double? Length)> collection = GetValues(mappingFromItem, filteredFaces, modelsFaceEncoding, normalizedRectangle);
        results = (from l in collection orderby l.Length select l).Take(1).ToArray();
        if (results.Any())
        {
            (Face _, double? length) = results.First();
            _Debug.Add(length);
        }
        return results;
    }

    private static List<(Face, double?)> GetMatchingFacesByFaceEncoding(Face[] filteredFaces, string? json)
    {
        List<(Face, double?)> results = new();
        string check;
        foreach (Face face in filteredFaces)
        {
            if (json is null || face.FaceEncoding is null)
                continue;
            if (!json.Contains(face.FaceEncoding.RawEncoding[0].ToString()))
                continue;
            check = JsonSerializer.Serialize(face.FaceEncoding);
            if (check != json)
                continue;
            results.Add(new(face, 0));
        }
        return results;
    }

    private static FileInfo? CheckFileThenGetFileInfo(string facesFileNameExtension, MappingFromItem mappingFromItem, string mappedFaceFile, List<(Face, double?)> checkFaces)
    {
        FileInfo? result = null;
        string checkFile;
        string? mappedFaceDirectory;
        string deterministicHashCodeKey;
        foreach ((Face face, _) in checkFaces)
        {
            if (checkFaces.Count != 1)
                break;
            if (face.FaceEncoding is null || face.Location is null || face.OutputResolution is null)
                throw new NotSupportedException();
            mappedFaceDirectory = Path.GetDirectoryName(mappedFaceFile);
            if (mappedFaceDirectory is null)
                throw new NotSupportedException();
            deterministicHashCodeKey = Shared.Models.Stateless.Methods.IMapping.GetDeterministicHashCodeKey(mappingFromItem.Id, face.Location, Shared.Models.Stateless.ILocation.Digits, face.OutputResolution);
            checkFile = Path.Combine(mappedFaceDirectory, $"{deterministicHashCodeKey}{mappingFromItem.ImageFileHolder.ExtensionLowered}{facesFileNameExtension}");
            if (checkFile == mappedFaceFile)
                continue;
            result = new FileInfo(checkFile);
        }
        return result;
    }

    private void AppendMatchingDuplicates(string mappedFaceFile, string[] matches)
    {
        string checkFile;
        FileInfo fileInfo = new(mappedFaceFile);
        List<(long Length, string FullName)> collection = new();
        if (fileInfo.Exists)
            collection.Add(new(fileInfo.Length, fileInfo.FullName));
        lock (_DuplicateMappedFaceFiles)
            _DuplicateMappedFaceFiles.Add(mappedFaceFile);
        foreach (string match in matches)
        {
            fileInfo = new(match);
            if (!fileInfo.Exists)
                continue;
            collection.Add(new(fileInfo.Length, fileInfo.FullName));
            break;
        }
        collection = collection.OrderBy(l => l.Length).ToList();
        for (int i = 0; i < collection.Count - 1; i++)
        {
            checkFile = string.Concat(collection[i].FullName, ".dup");
            if (File.Exists(checkFile))
                continue;
            File.Move(collection[i].FullName, checkFile);
        }
    }

    private static List<Face> FilterByIntersect(Face[] faces, int normalizedRectangle)
    {
        List<Face> results = new();
        bool useOldWay;
        double? percent;
        System.Drawing.Rectangle checkRectangle;
        System.Drawing.Rectangle sourceRectangle;
        System.Drawing.Rectangle intersectRectangle;
        foreach (Face face in faces)
        {
            if (face.Location is null || face.OutputResolution is null)
                continue;
            checkRectangle = new(face.Location.Left, face.Location.Top, face.Location.Right - face.Location.Left, face.Location.Bottom - face.Location.Top);
            for (int i = 1; i < 3; i++)
            {
                useOldWay = i == 1;
                sourceRectangle = Shared.Models.Stateless.Methods.ILocation.GetRectangle(checkRectangle, Shared.Models.Stateless.ILocation.Digits, Shared.Models.Stateless.ILocation.Factor, normalizedRectangle, face.OutputResolution, useOldWay);
                intersectRectangle = System.Drawing.Rectangle.Intersect(sourceRectangle, checkRectangle);
                if (intersectRectangle.Width == 0 || intersectRectangle.Height == 0)
                    continue;
                percent = (double)intersectRectangle.Width * intersectRectangle.Height / (checkRectangle.Width * checkRectangle.Height);
                if (percent < 0.000001)
                    continue;
                results.Add(face);
            }
        }
        return results;
    }

    public void LookForMatchFacesAndPossiblyRename(string facesFileNameExtension, string? eDistanceContentDirectory, MappingFromItem mappingFromItem, List<Face> faces, List<(string MappedFaceFile, int NormalizedRectangle, IReadOnlyList<MetadataExtractor.Directory>? Directories)> collection)
    {
        string? json;
        string[] matches;
        FileInfo? fileInfo;
        string mappedFaceFileName;
        List<(Face, double?)> checkFaces = new();
        Shared.Models.FaceEncoding? modelsFaceEncoding;
        Face[] filteredFaces = (from l in faces where l.FaceEncoding is not null && l.Location is not null && l.OutputResolution is not null select l).ToArray();
        if (filteredFaces.Length != faces.Count)
            checkFaces.Clear();
        foreach ((string mappedFaceFile, int normalizedRectangle, IReadOnlyList<MetadataExtractor.Directory>? directories) in collection)
        {
            if (!filteredFaces.Any())
                break;
            if (_Renamed.Contains(mappedFaceFile))
                continue;
            mappedFaceFileName = Path.GetFileName(mappedFaceFile);
            if (_DuplicateMappedFaceFiles.Contains(mappedFaceFileName))
                continue;
            checkFaces.Clear();
            if (directories is null)
                throw new NullReferenceException(nameof(directories));
            json = Metadata.Models.Stateless.IMetadata.GetFaceEncoding(directories);
            if (json is null)
            {
                if (!string.IsNullOrEmpty(eDistanceContentDirectory) && _DistanceMoveUnableToMatch)
                    MoveUnableToMatch(eDistanceContentDirectory, mappedFaceFile, mappedFaceFileName);
                continue;
            }
            checkFaces.AddRange(GetMatchingFacesByFaceEncoding(filteredFaces, json));
            if (checkFaces.Count == 1)
                _Debug.Add(0);
            if (checkFaces.Count != 1 && !string.IsNullOrEmpty(json))
            {
                checkFaces.Clear();
                modelsFaceEncoding = JsonSerializer.Deserialize<Shared.Models.FaceEncoding>(json);
                if (modelsFaceEncoding is null)
                    throw new NotSupportedException();
                checkFaces.AddRange(GetClosestFaceByDistanceIgnoringTolerance(mappingFromItem, normalizedRectangle, filteredFaces, modelsFaceEncoding));
            }
            if (!checkFaces.Any())
            {
                if (!string.IsNullOrEmpty(eDistanceContentDirectory) && _DistanceMoveUnableToMatch)
                    MoveUnableToMatch(eDistanceContentDirectory, mappedFaceFile, mappedFaceFileName);
                continue;
            }
            if (checkFaces.Count != 1)
            {
                if (!string.IsNullOrEmpty(eDistanceContentDirectory) && _DistanceMoveUnableToMatch)
                    MoveUnableToMatch(eDistanceContentDirectory, mappedFaceFile, mappedFaceFileName);
                continue;
            }
            fileInfo = CheckFileThenGetFileInfo(facesFileNameExtension, mappingFromItem, mappedFaceFile, checkFaces);
            if (fileInfo is not null)
            {
                if (_DistanceRenameToMatch && fileInfo is not null)
                {
                    if (fileInfo.Exists)
                        File.Delete(fileInfo.FullName);
                    File.Move(mappedFaceFile, fileInfo.FullName);
                    _Renamed.Add(mappedFaceFile);
                }
                continue;
            }
            if (_AllMappedFaceFileNames.Contains(mappedFaceFileName))
            {
                lock (_AllMappedFaceFiles)
                    matches = (from l in _AllMappedFaceFiles where l != mappedFaceFile && Path.GetFileName(l) == mappedFaceFileName select l).ToArray();
                if (matches.Any())
                    AppendMatchingDuplicates(mappedFaceFile, matches);
            }
            lock (_AllMappedFaceFiles)
                _AllMappedFaceFiles.Add(mappedFaceFile);
            lock (_AllMappedFaceFileNames)
                _AllMappedFaceFileNames.Add(mappedFaceFileName);
        }
    }

    public void Clear()
    {
        if (_Log is null)
            throw new NullReferenceException(nameof(_Log));
        double?[] debug = (from l in _Debug where l is null or not 0 select l).ToArray();
        if (debug.Any())
        {
            string debugMessage = $"{_Debug.Count - debug.Length} - {debug.Min()} - {_Debug.Max()}";
            _Log.Info(debugMessage);
        }
        if (_Moved.Any() || _Renamed.Any() || _DuplicateMappedFaceFiles.Any())
            throw new NotImplementedException("Restart!");
        _Debug.Clear();
        _Moved.Clear();
        _Renamed.Clear();
        _AllMappedFaceFiles.Clear();
        _AllMappedFaceFileNames.Clear();
        _DuplicateMappedFaceFiles.Clear();
    }

    public List<FaceDistanceContainer> GetMissingFaceDistanceContainer(int maxDegreeOfParallelism, long ticks, string dFacesCollectionDirectory, Dictionary<int, Dictionary<int, PersonContainer[]>> missingIdThenNormalizedRectangleToPersonContainers)
    {
        List<FaceDistanceContainer> results = new();
        string[] files;
        List<Face>? faces;
        int confidencePercent;
        int normalizedRectangle;
        bool? isWrongYear = null;
        FaceDistance faceDistance;
        List<(int id, string json)> collection = new();
        FaceDistanceContainer faceDistanceContainer;
        foreach (KeyValuePair<int, Dictionary<int, PersonContainer[]>> keyValuePair in missingIdThenNormalizedRectangleToPersonContainers)
        {
            files = Directory.GetFiles(dFacesCollectionDirectory, $"{keyValuePair.Key}*.json", SearchOption.TopDirectoryOnly);
            if (files.Length != 1)
                continue;
            collection.Add(new(keyValuePair.Key, Shared.Models.Stateless.Methods.IFace.GetJson(files[0])));
        }
        int totalSeconds = (int)Math.Floor(new TimeSpan(DateTime.Now.Ticks - ticks).TotalSeconds);
        ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = maxDegreeOfParallelism };
        string message = $") {collection.Count:000} Setting missing distance containers - {totalSeconds} total second(s)";
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        using ProgressBar progressBar = new(collection.Count, message, options);
        _ = Parallel.For(0, collection.Count, parallelOptions, (i, state) =>
        {
            progressBar.Tick();
            int id = collection[i].id;
            string json = collection[i].json;
            faces = JsonSerializer.Deserialize<List<Face>>(json);
            if (faces is null)
                throw new NullReferenceException(nameof(faces));
            foreach (Face face in faces)
            {
                if (face.FaceEncoding is null || face.Location is null || face.OutputResolution is null)
                    continue;
                confidencePercent = Shared.Models.Stateless.Methods.ILocation.GetConfidencePercent(_FaceConfidencePercent, _RangeFaceConfidence, face.Location.Confidence);
                normalizedRectangle = Shared.Models.Stateless.Methods.ILocation.GetNormalizedRectangle(face.Location, Shared.Models.Stateless.ILocation.Digits, face.OutputResolution);
                if (face.FaceDistance?.Encoding is not null && face.FaceDistance.Encoding is FaceRecognitionDotNet.FaceEncoding faceEncoding)
                    faceDistance = new(confidencePercent, faceEncoding, id, isWrongYear, face.DateTime, normalizedRectangle);
                else
                {
                    faceEncoding = FaceRecognition.LoadFaceEncoding(face.FaceEncoding.RawEncoding);
                    faceDistance = new(confidencePercent, faceEncoding, id, isWrongYear, face.DateTime, normalizedRectangle);
                    face.SetFaceDistance(faceDistance);
                }
                faceDistanceContainer = new(face, faceDistance);
                lock (results)
                    results.Add(faceDistanceContainer);
            }
        });
        return results;
    }

}